/*
								+---------------------------------+
								|                                 |
								|   ***   CTextureServer    ***   |
								|   ***   CTextureManager   ***   |
								|                                 |
								|  Copyright   -tHE SWINe- 2005  |
								|                                 |
								|        TextureManager.cpp       |
								|                                 |
								+---------------------------------+
*/

/*
 *
 *	03-06-2007
 *
 *	removed __declspec(dllexport)
 *
 */

#include <windows.h>
#include <crtdbg.h>
#include <malloc.h>
#include <stdio.h>
#include <GL/gl.h>
#include <GL/glext.h>
#include <math.h>
#include <time.h>
#include <vector>
#include "list.h"
#include "OpenGlDrv.h"
#include "OpenGl20.h"
#include "Glstate.h"
#include "Tga.h"
#include "Jpeg.h"
#include "TextureManager.h"

#include "callstack.h"

/*
 *		=== Saving bitmaps ===
 */

int Save_TrueColor_BMP(char *filename, TBmp *p_bmp)
{
	unsigned int n_scanline_size;
	unsigned char *p_buffer;
	BITMAPFILEHEADER bmfh;
	BITMAPINFOHEADER bmih;
	FILE *p_fw;
	int i, j;

	n_scanline_size = ((3 * p_bmp->n_width + 3) >> 2) << 2;
	if(!(p_buffer =  new unsigned char[n_scanline_size]))
		return false;
	// compute lenght of line-buffer

	if(!(p_fw = fopen(filename, "wb")))
		return false;
	// open file

	bmih.biSize = sizeof(BITMAPINFOHEADER);
	bmih.biWidth = p_bmp->n_width;
	bmih.biHeight = p_bmp->n_height;
	bmih.biPlanes = 1;
	bmih.biBitCount = 24;
	bmih.biCompression = BI_RGB;
	bmih.biSizeImage = p_bmp->n_width * p_bmp->n_height * 3;
	bmih.biXPelsPerMeter = 96;
	bmih.biYPelsPerMeter = 96;
	bmih.biClrUsed = 0;
	bmih.biClrImportant = 0;
	//
	bmfh.bfType = ('B' | ('M' << 8));
	bmfh.bfSize = sizeof(BITMAPFILEHEADER) + bmih.biSizeImage +
				  sizeof(BITMAPINFOHEADER);
	bmfh.bfReserved1 = 0;
	bmfh.bfReserved2 = 0;
	bmfh.bfOffBits = sizeof(BITMAPFILEHEADER) +
					 sizeof(BITMAPINFOHEADER);
	// create header

	fwrite(&bmfh, sizeof(BITMAPFILEHEADER), 1, p_fw);
	fwrite(&bmih, sizeof(BITMAPINFOHEADER), 1, p_fw);
	// writing header

	for(i = p_bmp->n_height - 1; i >= 0; i --) {
		for(j = 0; j < p_bmp->n_width; j ++) {
			p_buffer[j * 3] = (char)((p_bmp->p_buffer[i * p_bmp->n_width + j] >> 16) & 0xff);
			p_buffer[j * 3 + 1] = (char)((p_bmp->p_buffer[i * p_bmp->n_width + j] >> 8) & 0xff);
			p_buffer[j * 3 + 2] = (char)(p_bmp->p_buffer[i * p_bmp->n_width + j] & 0xff);
			// saved in BGR, not RGB
		}

		if(fwrite(p_buffer, 1, n_scanline_size, p_fw) != n_scanline_size) {
			delete[] p_buffer;
			fclose(p_fw);

			return false;
		}
	}

	fclose(p_fw);
	delete[] p_buffer;

	return true;
}

/*
 *								=== CTextureServer ===
 */

CTextureServer::CTextureServer()
	:m_b_sorted(true)
{
	__FuncGuard("CTextureServer::CTextureServer");
}

CTextureServer::~CTextureServer()
{
	__FuncGuard("CTextureServer::~CTextureServer");

	if(m_texture_list.size())
		Delete_AllTextures();
}

void CTextureServer::Delete_AllTextures()
{
	__FuncGuard("CTextureServer::Delete_AllTextures");

	int i;
	
	for(i = 0; i < m_texture_list.size(); i ++) {
		glDeleteTextures(1, &m_texture_list[i].n_open_gl_id);
		//
		if(m_texture_list[i].p_s_filename)
			delete[] m_texture_list[i].p_s_filename;
		if(m_texture_list[i].p_s_alpha_filename)
			delete[] m_texture_list[i].p_s_alpha_filename;
	}
	m_texture_list.clear();
}

int CTextureServer::Load_Bitmaps(char *p_s_filename, char *p_s_alpha_filename, TBmp &r_t_rgba_result)
{
	__FuncGuard("CTextureServer::Load_Bitmaps");

	TBmp *p_bitmap, *p_alpha_bitmap;

	if(!(p_bitmap = p_Load_TGA(p_s_filename)) &&
	   !(p_bitmap = p_Load_JPG(p_s_filename)))
		return 0;
	// nahraje bitmapu

	if(p_s_alpha_filename &&
	   !(p_alpha_bitmap = p_Load_TGA(p_s_alpha_filename)) &&
	   !(p_bitmap = p_Load_JPG(p_s_alpha_filename))) {
		delete[] p_bitmap->p_buffer;
		delete p_bitmap;
		return 0;
	}
	// nahraje alpha-bitmapu

	if(p_s_alpha_filename) {
		if(p_bitmap->n_width != p_alpha_bitmap->n_width ||
		   p_bitmap->n_height != p_alpha_bitmap->n_height) {
			delete[] p_bitmap->p_buffer;
			delete[] p_alpha_bitmap->p_buffer;
			delete p_bitmap;
			delete p_alpha_bitmap;
			return 0;
		}
		// mus mt stejn rozmry ...

		if(p_alpha_bitmap->b_alpha) {
			for(int i = 0; i < p_bitmap->n_width * p_bitmap->n_height; i ++) {
				p_bitmap->p_buffer[i] &= 0xffffff;
				p_bitmap->p_buffer[i] |= p_alpha_bitmap->p_buffer[i] & 0xff000000;
			}
		} else {
			for(int i = 0; i < p_bitmap->n_width * p_bitmap->n_height; i ++) {
				p_bitmap->p_buffer[i] &= 0xffffff;
				p_bitmap->p_buffer[i] |= (p_alpha_bitmap->p_buffer[i] & 0xff0000) << 8;
			}
		}
		// pokud m alfu, vezme alfu. jinak vezme erven kanl

		delete[] p_alpha_bitmap->p_buffer;
		delete p_alpha_bitmap;
		// tohle u nepotebujeme

		p_bitmap->b_alpha = true;
		// to jsme tam prv pidali ...
	}
	// pimixuje alphu z druh bitmapy, pokud je zadan

	r_t_rgba_result = *p_bitmap;
	delete p_bitmap;
	// vrt zbytek

	return true;
}

unsigned int CTextureServer::Load_Texture_1D(CGLState *p_state,
	char *p_s_filename, GLenum n_format, int b_create_mipmaps, char *p_s_alpha_filename)
{
	__FuncGuard("CTextureServer::Load_Texture_1D");

	TTextureInfo t_tex_info;
	TBmp t_bitmap;

	if(!Load_Bitmaps(p_s_filename, p_s_alpha_filename, t_bitmap))
		return 0;
	if(!t_tex_info.SetInfo(p_s_filename, p_s_alpha_filename, GL_TEXTURE_1D, t_bitmap.n_width, 1, 1)) {
		delete[] t_bitmap.p_buffer;
		return 0;
	}

	p_state->EnableTexture1D();
	p_state->DisableTexture2D();
	p_state->DisableTexture3D();
	p_state->DisableTextureCubeMap();

	glGenTextures(1, &t_tex_info.n_open_gl_id);
	p_state->BindTexture1D(t_tex_info.n_open_gl_id);
	// vytvo texturu a aktivujue ji

    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	// pole bez mezer

	if(b_create_mipmaps)
		glTexParameteri(GL_TEXTURE_1D, GL_GENERATE_MIPMAP, GL_TRUE);

	glTexImage1D(GL_TEXTURE_1D, 0, n_format,
		t_bitmap.n_width, 0, GL_RGBA, GL_UNSIGNED_BYTE, t_bitmap.p_buffer);
	// nahraje data

	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER,
		(b_create_mipmaps)? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR);
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	// zatm nemme mipmapy

	delete[] t_bitmap.p_buffer;
	// tohle u nen poteba

	if(glGetError() != GL_NO_ERROR) {
		t_tex_info.DeleteTexture();
		return 0;
	}
	// zkontroluje jestli nedola pam

	if(m_texture_list.size() && m_texture_list.back().n_open_gl_id > t_tex_info.n_open_gl_id)
		m_b_sorted = false;
	m_texture_list.push_back(t_tex_info);
	if(!m_texture_list.size()) {
		t_tex_info.DeleteTexture();
		return 0;
	}
	// vraz do seznamu textur

	return t_tex_info.n_open_gl_id;
}

unsigned int CTextureServer::Load_Texture_2D(CGLState *p_state,
	char *p_s_filename, GLenum n_format, int b_create_mipmaps, char *p_s_alpha_filename)
{
	__FuncGuard("CTextureServer::Load_Texture_2D");

	TTextureInfo t_tex_info;
	TBmp t_bitmap;

	if(!Load_Bitmaps(p_s_filename, p_s_alpha_filename, t_bitmap))
		return 0;
	if(!t_tex_info.SetInfo(p_s_filename, p_s_alpha_filename, GL_TEXTURE_2D,
	   t_bitmap.n_width, t_bitmap.n_height, 1)) {
		delete[] t_bitmap.p_buffer;
		return 0;
	}

	p_state->DisableTexture1D();
	p_state->EnableTexture2D();
	p_state->DisableTexture3D();
	p_state->DisableTextureCubeMap();

	glGenTextures(1, &t_tex_info.n_open_gl_id);
	p_state->BindTexture2D(t_tex_info.n_open_gl_id);
	// vytvo texturu a aktivujue ji

    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	// pole bez mezer

	if(b_create_mipmaps)
		glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
		(b_create_mipmaps)? GL_LINEAR_MIPMAP_LINEAR : GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	// zatm nemme mipmapy

	glTexImage2D(GL_TEXTURE_2D, 0, n_format,
		t_bitmap.n_width, t_bitmap.n_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, t_bitmap.p_buffer);
	// nahraje data

	delete[] t_bitmap.p_buffer;
	// tohle u nen poteba

	if(glGetError() != GL_NO_ERROR) {
		t_tex_info.DeleteTexture();
		return 0;
	}
	// zkontroluje jestli nedola pam

	if(m_texture_list.size() && m_texture_list.back().n_open_gl_id > t_tex_info.n_open_gl_id)
		m_b_sorted = false;
	m_texture_list.push_back(t_tex_info);
	if(!m_texture_list.size()) {
		t_tex_info.DeleteTexture();
		return 0;
	}
	// vraz do seznamu textur

	return t_tex_info.n_open_gl_id;
}

unsigned int CTextureServer::Load_Texture_3D(CGLState *p_state,
	char *p_s_filename, int n_depth, GLenum n_format, int b_create_mipmaps, char *p_s_alpha_filename)
{
	__FuncGuard("CTextureServer::Load_Texture_3D");

	TTextureInfo t_tex_info;
	TBmp t_bitmap;

	if(!t_tex_info.SetInfo(p_s_filename, p_s_alpha_filename, GL_TEXTURE_3D, 1, 1, n_depth))
		return 0;

	p_state->DisableTexture1D();
	p_state->DisableTexture2D();
	p_state->EnableTexture3D();
	p_state->DisableTextureCubeMap();

	glGenTextures(1, &t_tex_info.n_open_gl_id);
	p_state->BindTexture3D(t_tex_info.n_open_gl_id);
	// vytvo texturu a aktivujue ji

    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
	// pole bez mezer

	if(b_create_mipmaps)
		glTexParameteri(GL_TEXTURE_3D, GL_GENERATE_MIPMAP, GL_TRUE);

	for(int i = 0; i < n_depth; i ++) {
		if(p_s_alpha_filename) {
			char p_s_buffer[4096];
			char p_s_buffer_alpha[4096];

			sprintf(p_s_buffer, p_s_filename, i);
			sprintf(p_s_buffer_alpha, p_s_alpha_filename, i);
			if(!Load_Bitmaps(p_s_buffer, p_s_buffer_alpha, t_bitmap)) {
				t_tex_info.DeleteTexture();
				return 0;
			}
		} else {
			char p_s_buffer[4096];

			sprintf(p_s_buffer, p_s_filename, i);
			if(!Load_Bitmaps(p_s_buffer, 0, t_bitmap)) {
				t_tex_info.DeleteTexture();
				return 0;
			}
		}
		// zkus nahrt bitmapy

		if(!i) {
			glTexImage3D(GL_TEXTURE_3D, 0, n_format,
				t_bitmap.n_width, t_bitmap.n_height, n_depth, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
			// vytvo texturu ...

			t_tex_info.n_size[0] = t_bitmap.n_width;
			t_tex_info.n_size[1] = t_bitmap.n_height;
		}

		glTexSubImage3D(GL_TEXTURE_3D, 0, 0, 0, i, t_bitmap.n_width, t_bitmap.n_height,
			1, GL_RGBA, GL_UNSIGNED_BYTE, t_bitmap.p_buffer);
		// nahraje jeden "pltek" 3d textury

		delete[] t_bitmap.p_buffer;
		// tohle u nen poteba
	}

	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER,
		(b_create_mipmaps)? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	// zatm nemme mipmapy

	if(glGetError() != GL_NO_ERROR) {
		t_tex_info.DeleteTexture();
		return 0;
	}
	// zkontroluje jestli nedola pam

	if(m_texture_list.size() && m_texture_list.back().n_open_gl_id > t_tex_info.n_open_gl_id)
		m_b_sorted = false;
	m_texture_list.push_back(t_tex_info);
	if(!m_texture_list.size()) {
		t_tex_info.DeleteTexture();
		return 0;
	}
	// vraz do seznamu textur

	return t_tex_info.n_open_gl_id;
}

static GLenum p_cube_face[6] = {
	GL_TEXTURE_CUBE_MAP_POSITIVE_X,
	GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
	GL_TEXTURE_CUBE_MAP_POSITIVE_Y,
	GL_TEXTURE_CUBE_MAP_NEGATIVE_Y,
	GL_TEXTURE_CUBE_MAP_POSITIVE_Z,
	GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
};

static char p_s_cube_face[6][5] = {"posx", "negx", "posy", "negy", "posz", "negz"};

unsigned int CTextureServer::Load_Texture_CubeMap(CGLState *p_state,
	char *p_s_filename, GLenum n_format, int b_create_mipmaps, char *p_s_alpha_filename)
{
	__FuncGuard("CTextureServer::Load_Texture_CubeMap");

	TTextureInfo t_tex_info;
	TBmp t_bitmap;

	if(!t_tex_info.SetInfo(p_s_filename, p_s_alpha_filename, GL_TEXTURE_CUBE_MAP, 1, 1, 1))
		return 0;

	p_state->DisableTexture1D();
	p_state->DisableTexture2D();
	p_state->DisableTexture3D();
	p_state->EnableTextureCubeMap();

	glGenTextures(1, &t_tex_info.n_open_gl_id);
	p_state->BindTextureCubeMap(t_tex_info.n_open_gl_id);
	// vytvo texturu a aktivujue ji

	if(b_create_mipmaps)
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_GENERATE_MIPMAP, GL_TRUE);

    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	// pole bez mezer

	for(int i = 0; i < 6; i ++) {
		if(p_s_alpha_filename) {
			char p_s_buffer[4096];
			char p_s_buffer_alpha[4096];

			sprintf(p_s_buffer, p_s_filename, p_s_cube_face[i]);
			sprintf(p_s_buffer_alpha, p_s_alpha_filename, p_s_cube_face[i]);
			if(!Load_Bitmaps(p_s_buffer, p_s_buffer_alpha, t_bitmap)) {
				t_tex_info.DeleteTexture();
				return 0;
			}
		} else {
			char p_s_buffer[4096];

			sprintf(p_s_buffer, p_s_filename, p_s_cube_face[i]);
			if(!Load_Bitmaps(p_s_buffer, 0, t_bitmap)) {
				t_tex_info.DeleteTexture();
				return 0;
			}
		}
		// zkus nahrt bitmapy

		glTexImage2D(p_cube_face[i], 0, n_format,
			t_bitmap.n_width, t_bitmap.n_height,
			0, GL_RGBA, GL_UNSIGNED_BYTE, t_bitmap.p_buffer);
		// nahraje stnu cubemapy

		if(p_cube_face[i] == GL_TEXTURE_CUBE_MAP_NEGATIVE_Z) {
			t_tex_info.n_size[0] = t_bitmap.n_width;
			t_tex_info.n_size[1] = t_bitmap.n_height;
		} else if(p_cube_face[i] == GL_TEXTURE_CUBE_MAP_POSITIVE_Y)
			t_tex_info.n_size[2] = t_bitmap.n_height;
		// nahraje rozmry cubemapy

		delete[] t_bitmap.p_buffer;
		// tohle u nen poteba
	}

	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER,
		(b_create_mipmaps)? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR);
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	// zatm nemme mipmapy

	if(glGetError() != GL_NO_ERROR) {
		t_tex_info.DeleteTexture();
		return 0;
	}
	// zkontroluje jestli nedola pam

	if(m_texture_list.size() && m_texture_list.back().n_open_gl_id > t_tex_info.n_open_gl_id)
		m_b_sorted = false;
	m_texture_list.push_back(t_tex_info);
	if(!m_texture_list.size()) {
		t_tex_info.DeleteTexture();
		return 0;
	}
	// vraz do seznamu textur

	return t_tex_info.n_open_gl_id;
}

unsigned int CTextureServer::Create_Texture_1D(CGLState *p_state,
	int n_width, GLenum n_format, int b_create_mipmaps, void *p_data,
	GLenum n_data_format, GLenum n_data_type)
{
	__FuncGuard("CTextureServer::Create_Texture_1D");

	TTextureInfo t_tex_info;

	if(!t_tex_info.SetInfo(0, 0, GL_TEXTURE_1D, n_width, 1, 1))
		return 0;

	p_state->EnableTexture1D();
	p_state->DisableTexture2D();
	p_state->DisableTexture3D();
	p_state->DisableTextureCubeMap();

	glGenTextures(1, &t_tex_info.n_open_gl_id);
	p_state->BindTexture1D(t_tex_info.n_open_gl_id);
	// vytvo texturu a aktivujue ji

	if(b_create_mipmaps)
		glTexParameteri(GL_TEXTURE_1D, GL_GENERATE_MIPMAP, GL_TRUE);

	glTexImage1D(GL_TEXTURE_1D, 0, n_format,
		n_width, 0, n_data_format, n_data_type, p_data);
	// nahraje data

	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER,
		(b_create_mipmaps)? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR);
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	// zatm nemme mipmapy

	if(glGetError() != GL_NO_ERROR) {
		t_tex_info.DeleteTexture();
		return 0;
	}
	// zkontroluje jestli nedola pam

	if(m_texture_list.size() && m_texture_list.back().n_open_gl_id > t_tex_info.n_open_gl_id)
		m_b_sorted = false;
	m_texture_list.push_back(t_tex_info);
	if(!m_texture_list.size()) {
		t_tex_info.DeleteTexture();
		return 0;
	}
	// vraz do seznamu textur

	return t_tex_info.n_open_gl_id;
}

unsigned int CTextureServer::Create_Texture_2D(CGLState *p_state,
	int n_width, int n_height, GLenum n_format, int b_create_mipmaps,
	void *p_data, GLenum n_data_format, GLenum n_data_type)
{
	__FuncGuard("CTextureServer::Create_Texture_2D");

	TTextureInfo t_tex_info;

	if(!t_tex_info.SetInfo(0, 0, GL_TEXTURE_2D, n_width, n_height, 1))
		return 0;

	p_state->DisableTexture1D();
	p_state->EnableTexture2D();
	p_state->DisableTexture3D();
	p_state->DisableTextureCubeMap();

	// nahraje data
	glGenTextures(1, &t_tex_info.n_open_gl_id);
	p_state->BindTexture2D(t_tex_info.n_open_gl_id);
	// vytvo texturu a aktivujue ji

	if(b_create_mipmaps)
		glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
		(b_create_mipmaps)? GL_LINEAR_MIPMAP_LINEAR : GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	// zatm nemme mipmapy

	glTexImage2D(GL_TEXTURE_2D, 0, n_format,
		n_width, n_height, 0, n_data_format, n_data_type, p_data);
	// nahraje data

	if(glGetError() != GL_NO_ERROR) {
		t_tex_info.DeleteTexture();
		return 0;
	}
	// zkontroluje jestli nedola pam

	if(m_texture_list.size() && m_texture_list.back().n_open_gl_id > t_tex_info.n_open_gl_id)
		m_b_sorted = false;
	m_texture_list.push_back(t_tex_info);
	if(!m_texture_list.size()) {
		t_tex_info.DeleteTexture();
		return 0;
	}
	// vraz do seznamu textur

	return t_tex_info.n_open_gl_id;
}

unsigned int CTextureServer::Create_Texture_3D(CGLState *p_state,
	int n_width, int n_height, int n_depth, GLenum n_format, int b_create_mipmaps,
	void *p_data, GLenum n_data_format, GLenum n_data_type)
{
	__FuncGuard("CTextureServer::Create_Texture_3D");

	TTextureInfo t_tex_info;

	if(!t_tex_info.SetInfo(0, 0, GL_TEXTURE_3D, n_width, n_height, n_depth))
		return 0;

	p_state->DisableTexture1D();
	p_state->DisableTexture2D();
	p_state->EnableTexture3D();
	p_state->DisableTextureCubeMap();

	glGenTextures(1, &t_tex_info.n_open_gl_id);
	p_state->BindTexture3D(t_tex_info.n_open_gl_id);
	// vytvo texturu a aktivujue ji

	if(b_create_mipmaps)
		glTexParameteri(GL_TEXTURE_3D, GL_GENERATE_MIPMAP, GL_TRUE);

	glTexImage3D(GL_TEXTURE_3D, 0, n_format,
		n_width, n_height, n_depth, 0, n_data_format, n_data_type, p_data);
	// nahraje data

	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER,
		(b_create_mipmaps)? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	// zatm nemme mipmapy

	if(glGetError() != GL_NO_ERROR) {
		t_tex_info.DeleteTexture();
		return 0;
	}
	// zkontroluje jestli nedola pam

	if(m_texture_list.size() && m_texture_list.back().n_open_gl_id > t_tex_info.n_open_gl_id)
		m_b_sorted = false;
	m_texture_list.push_back(t_tex_info);
	if(!m_texture_list.size()) {
		t_tex_info.DeleteTexture();
		return 0;
	}
	// vraz do seznamu textur

	return t_tex_info.n_open_gl_id;
}

unsigned int CTextureServer::Create_Texture_CubeMap(CGLState *p_state,
	int n_width, GLenum n_format, int b_create_mipmaps)
{
	__FuncGuard("CTextureServer::Create_Texture_CubeMap");

	TTextureInfo t_tex_info;

	if(!t_tex_info.SetInfo(0, 0, GL_TEXTURE_CUBE_MAP, n_width, n_width, n_width))
		return 0;

	p_state->DisableTexture1D();
	p_state->DisableTexture2D();
	p_state->DisableTexture3D();
	p_state->EnableTextureCubeMap();

	glGenTextures(1, &t_tex_info.n_open_gl_id);
	p_state->BindTextureCubeMap(t_tex_info.n_open_gl_id);
	// vytvo texturu a aktivujue ji

	if(b_create_mipmaps)
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_GENERATE_MIPMAP, GL_TRUE);

	for(int i = 0; i < 6; i ++) {
		glTexImage2D(p_cube_face[i], 0, n_format,
			n_width, n_width, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
	}
	// nahraje data

	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER,
		(b_create_mipmaps)? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR);
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	// zatm nemme mipmapy

	if(glGetError() != GL_NO_ERROR) {
		t_tex_info.DeleteTexture();
		return 0;
	}
	// zkontroluje jestli nedola pam

	if(m_texture_list.size() && m_texture_list.back().n_open_gl_id > t_tex_info.n_open_gl_id)
		m_b_sorted = false;
	m_texture_list.push_back(t_tex_info);
	if(!m_texture_list.size()) {
		t_tex_info.DeleteTexture();
		return 0;
	}
	// vraz do seznamu textur

	return t_tex_info.n_open_gl_id;
}

int CTextureServer::n_CompareTexInfo(const void *p_elem1, const void *p_elem2)
{
	__FuncGuard("CTextureServer::n_CompareTexInfo");

	return ((TTextureInfo*)p_elem1)->n_open_gl_id -
		((TTextureInfo*)p_elem2)->n_open_gl_id;
}

void CTextureServer::SortTextures()
{
	__FuncGuard("CTextureServer::SortTextures");

	if(m_b_sorted)
		return;
	m_b_sorted = true;
	qsort(m_texture_list.begin(), m_texture_list.size(),
		sizeof(TTextureInfo), n_CompareTexInfo);
}

int CTextureServer::n_Texture_Size(unsigned int n_ogl_id, int n_dimension) const
{
	__FuncGuard("CTextureServer::n_Texture_Size");

	if(m_b_sorted) {
		TTextureInfo t_temp, *p_found;
		t_temp.n_open_gl_id = n_ogl_id;
		p_found = (TTextureInfo*)bsearch(&t_temp, m_texture_list.begin(), m_texture_list.size(), sizeof(TTextureInfo),
			n_CompareTexInfo);
		if(p_found)
			return p_found->n_size[n_dimension];
	} else {
		for(int i = 0; i < m_texture_list.size(); i ++) {
			if(m_texture_list[i].n_open_gl_id == n_ogl_id)
				return m_texture_list[i].n_size[n_dimension];
		}
	}
	return -1;
}

unsigned int CTextureServer::Find_Texture(char *p_s_filename, unsigned int n_prev_texture, char *p_s_alpha_filename) const
{
	__FuncGuard("CTextureServer::Find_Texture");

	int b_yet = !n_prev_texture;
	for(int i = 0; i < m_texture_list.size(); i ++) {
		if(m_texture_list[i].n_open_gl_id == n_prev_texture) {
			b_yet = true;
			continue;
		}
		if(!b_yet)
			continue;
		if(stricmp(m_texture_list[i].p_s_filename, p_s_filename))
			continue;
		if(!p_s_alpha_filename != !m_texture_list[i].p_s_alpha_filename ||
		   stricmp(m_texture_list[i].p_s_alpha_filename, p_s_alpha_filename))
			continue;

		return m_texture_list[i].n_open_gl_id;
	}
	return 0;
}

int CTextureServer::Use_Texture(unsigned int n_ogl_id)
{
	__FuncGuard("CTextureServer::Use_Texture");

	if(m_b_sorted) {
		TTextureInfo t_temp, *p_found;
		t_temp.n_open_gl_id = n_ogl_id;
		p_found = (TTextureInfo*)bsearch(&t_temp, m_texture_list.begin(), m_texture_list.size(), sizeof(TTextureInfo),
			n_CompareTexInfo);
		if(p_found)
			return p_found->n_references ++;
	} else {
		for(int i = 0; i < m_texture_list.size(); i ++) {
			if(m_texture_list[i].n_open_gl_id == n_ogl_id) {
				m_texture_list[i].n_references ++;
				return true;
			}
		}
	}
	return false;
}

int CTextureServer::Delete_Texture(unsigned int n_ogl_id)
{
	__FuncGuard("CTextureServer::Delete_Texture");

	if(m_b_sorted) {
		TTextureInfo t_temp, *p_found;
		t_temp.n_open_gl_id = n_ogl_id;
		p_found = (TTextureInfo*)bsearch(&t_temp, m_texture_list.begin(), m_texture_list.size(), sizeof(TTextureInfo),
			n_CompareTexInfo);
		if(p_found) {
			if(!(-- p_found->n_references)) {
				p_found->DeleteTexture();
				m_texture_list.erase(p_found);
			}
			return true;
		}
	} else {
		for(int i = 0; i < m_texture_list.size(); i ++) {
			if(m_texture_list[i].n_open_gl_id == n_ogl_id) {
				if(!(-- m_texture_list[i].n_references)) {
					m_texture_list[i].DeleteTexture();
					m_texture_list.erase(&m_texture_list[i]);
				}
				return true;
			}
		}
	}

	return false;
}

/*
 *								=== ~CTextureServer ===
 */

/*
 *								=== CTextureManager ===
 */

CTextureServer *CTextureManager::m_p_texture_server = 0;
int CTextureManager::m_n_instances = 0;

CTextureManager::CTextureManager(CGLState *p_state)
	:m_p_state(p_state)
{
	__FuncGuard("CTextureManager::CTextureManager");

	_ASSERTE(p_state);
	if(!m_p_texture_server)
		m_p_texture_server = new CTextureServer;
	m_n_instances ++;
}

CTextureManager::~CTextureManager()
{
	__FuncGuard("CTextureManager::~CTextureManager");

	if(!(-- m_n_instances)) {
		delete m_p_texture_server;
		m_p_texture_server = 0;
	}
	m_local_texture_list.clear();
}

unsigned int CTextureManager::Load_Texture_1D(char *p_s_filename,
	GLenum n_format, int b_create_mipmaps, char *p_s_alpha_filename)
{
	__FuncGuard("CTextureManager::Load_Texture_1D");

	unsigned int n_tex_id;

	_ASSERTE(m_p_state);
	_ASSERTE(m_p_texture_server);

	n_tex_id = m_p_texture_server->Load_Texture_1D(m_p_state,
		p_s_filename, n_format, b_create_mipmaps, p_s_alpha_filename);

	if(n_tex_id)
		m_local_texture_list.push_back(n_tex_id);
	// vlo do loklnho seznamu ...

	return n_tex_id;
}

unsigned int CTextureManager::Load_Texture_2D(char *p_s_filename,
	GLenum n_format, int b_create_mipmaps, char *p_s_alpha_filename)
{
	__FuncGuard("CTextureManager::Load_Texture_2D");

	unsigned int n_tex_id;

	_ASSERTE(m_p_state);
	_ASSERTE(m_p_texture_server);

	n_tex_id = m_p_texture_server->Load_Texture_2D(m_p_state,
		p_s_filename, n_format, b_create_mipmaps, p_s_alpha_filename);

	if(n_tex_id)
		m_local_texture_list.push_back(n_tex_id);
	// vlo do loklnho seznamu ...

	return n_tex_id;
}

unsigned int CTextureManager::Load_Texture_3D(char *p_s_filename,
	int n_depth, GLenum n_format, int b_create_mipmaps, char *p_s_alpha_filename)
{
	__FuncGuard("CTextureManager::Load_Texture_3D");

	unsigned int n_tex_id;

	_ASSERTE(m_p_state);
	_ASSERTE(m_p_texture_server);

	n_tex_id = m_p_texture_server->Load_Texture_3D(m_p_state,
		p_s_filename, n_depth, n_format, b_create_mipmaps, p_s_alpha_filename);

	if(n_tex_id)
		m_local_texture_list.push_back(n_tex_id);
	// vlo do loklnho seznamu ...

	return n_tex_id;
}

unsigned int CTextureManager::Load_Texture_CubeMap(char *p_s_filename,
	GLenum n_format, int b_create_mipmaps, char *p_s_alpha_filename)
{
	__FuncGuard("CTextureManager::Load_Texture_CubeMap");

	unsigned int n_tex_id;

	_ASSERTE(m_p_state);
	_ASSERTE(m_p_texture_server);

	n_tex_id = m_p_texture_server->Load_Texture_CubeMap(m_p_state,
		p_s_filename, n_format, b_create_mipmaps, p_s_alpha_filename);

	if(n_tex_id)
		m_local_texture_list.push_back(n_tex_id);
	// vlo do loklnho seznamu ...

	return n_tex_id;
}


unsigned int CTextureManager::Create_Texture_1D(int n_width, GLenum n_format,
	int b_create_mipmaps, void *p_data, GLenum n_data_format, GLenum n_data_type)
{
	__FuncGuard("CTextureManager::Create_Texture_1D");

	unsigned int n_tex_id;

	_ASSERTE(m_p_state);
	_ASSERTE(m_p_texture_server);

	n_tex_id = m_p_texture_server->Create_Texture_1D(m_p_state,
		n_width, n_format, b_create_mipmaps, p_data, n_data_format, n_data_type);

	if(n_tex_id)
		m_local_texture_list.push_back(n_tex_id);
	// vlo do loklnho seznamu ...

	return n_tex_id;
}

unsigned int CTextureManager::Create_Texture_2D(int n_width, int n_height,
	GLenum n_format, int b_create_mipmaps,
	void *p_data, GLenum n_data_format, GLenum n_data_type)
{
	__FuncGuard("CTextureManager::Create_Texture_2D");

	unsigned int n_tex_id;

	_ASSERTE(m_p_state);
	_ASSERTE(m_p_texture_server);

	n_tex_id = m_p_texture_server->Create_Texture_2D(m_p_state,
		n_width, n_height, n_format, b_create_mipmaps, p_data, n_data_format, n_data_type);

	if(n_tex_id)
		m_local_texture_list.push_back(n_tex_id);
	// vlo do loklnho seznamu ...

	return n_tex_id;
}

unsigned int CTextureManager::Create_Texture_3D(int n_width, int n_height,
	int n_depth, GLenum n_format, int b_create_mipmaps,
	void *p_data, GLenum n_data_format, GLenum n_data_type)
{
	__FuncGuard("CTextureManager::Create_Texture_3D");

	unsigned int n_tex_id;

	_ASSERTE(m_p_state);
	_ASSERTE(m_p_texture_server);

	n_tex_id = m_p_texture_server->Create_Texture_3D(m_p_state,
		n_width, n_height, n_depth, n_format, b_create_mipmaps, p_data, n_data_format, n_data_type);

	if(n_tex_id)
		m_local_texture_list.push_back(n_tex_id);
	// vlo do loklnho seznamu ...

	return n_tex_id;
}

unsigned int CTextureManager::Create_Texture_CubeMap(int n_width,
	GLenum n_format, int b_create_mipmaps)
{
	__FuncGuard("CTextureManager::Create_Texture_CubeMap");

	unsigned int n_tex_id;

	_ASSERTE(m_p_state);
	_ASSERTE(m_p_texture_server);

	n_tex_id = m_p_texture_server->Create_Texture_CubeMap(m_p_state,
		n_width, n_format, b_create_mipmaps);

	if(n_tex_id)
		m_local_texture_list.push_back(n_tex_id);
	// vlo do loklnho seznamu ...

	return n_tex_id;
}

unsigned int CTextureManager::Find_Texture(char *p_s_filename, unsigned int n_prev_texture, char *p_s_alpha_filename)
{
	__FuncGuard("CTextureManager::Find_Texture");

	_ASSERTE(m_p_state);
	_ASSERTE(m_p_texture_server);

	return m_p_texture_server->Find_Texture(p_s_filename, n_prev_texture, p_s_alpha_filename);
}

void CTextureManager::SortTextures()
{
	__FuncGuard("CTextureManager::SortTextures");

	_ASSERTE(m_p_state);
	_ASSERTE(m_p_texture_server);

	m_p_texture_server->SortTextures();
}

int CTextureManager::n_Texture_Size(unsigned int n_ogl_id, int n_dimension)
{
	__FuncGuard("CTextureManager::n_Texture_Size");

	_ASSERTE(m_p_state);
	_ASSERTE(m_p_texture_server);

	return m_p_texture_server->n_Texture_Size(n_ogl_id, n_dimension);
}

int CTextureManager::Use_Texture(unsigned int n_ogl_id)
{
	__FuncGuard("CTextureManager::Use_Texture");

	_ASSERTE(m_p_state);
	_ASSERTE(m_p_texture_server);

	return m_p_texture_server->Use_Texture(n_ogl_id);
}

int CTextureManager::Delete_Texture(unsigned int n_ogl_id)
{
	__FuncGuard("CTextureManager::Delete_Texture");

	_ASSERTE(m_p_state);
	_ASSERTE(m_p_texture_server);

	for(int i = 0; i < m_local_texture_list.size(); i ++) {
		if(m_local_texture_list[i] == n_ogl_id)
			m_local_texture_list.erase(&m_local_texture_list[i --]);
	}
	m_p_texture_server->Delete_Texture(n_ogl_id);

	return 1;
}

int CTextureManager::Delete_LocalTextures()
{
	__FuncGuard("CTextureManager::Delete_LocalTextures");

	int i;

	_ASSERTE(m_p_state);
	_ASSERTE(m_p_texture_server);

	for(i = 0; i < m_local_texture_list.size(); i ++)
		m_p_texture_server->Delete_Texture(m_local_texture_list[i]);
	m_local_texture_list.clear();

	return i;
}

const CTextureServer *CTextureManager::p_Server()
{
	__FuncGuard("CTextureManager::p_Server");

	return m_p_texture_server;
}

/*
 *								=== ~CTextureManager ===
 */