/*
								+---------------------------------+
								|                                 |
								|     ***   CTextureMan   ***     |
								|                                 |
								|  Copyright   -tHE SWINe- 2002  |
								|                                 |
								|         TextureManager.h        |
								|                                 |
								+---------------------------------+
*/

/*
 *
 *	03-06-2007
 *
 *	removed __declspec(dllexport)
 *
 */

#ifndef __TEXTUREMAN_INCLUDED
#define __TEXTUREMAN_INCLUDED

#if defined(_MSC_VER) && !defined(__MWERKS__)
#pragma warning(4:4251)
// switch-off warning c4251 ("this class needs to have dll interface to be used by other calss clients")
#endif

struct /*__declspec(dllexport)*/ TTextureInfo {
	char *p_s_filename;
	char *p_s_alpha_filename;
	//
	int n_references;
	//
	unsigned int n_open_gl_id;
	int n_texture_type;
	//
	int n_size[3];

	TTextureInfo()
		:p_s_filename(0), 
		p_s_alpha_filename(0),
		n_references(0),
		n_open_gl_id(0),
		n_texture_type(GL_TEXTURE_2D)
	{
		memset(n_size, 0, sizeof(n_size));
	}

	void DeleteTexture()
	{
		if(p_s_filename)
			delete[] p_s_filename; // me bt nula pro on-the-fly vytven textury
		if(p_s_alpha_filename)
			delete[] p_s_alpha_filename;
		if(n_open_gl_id)
			glDeleteTextures(1, &n_open_gl_id);
	}

	int SetInfo(char *p_s_new_filename, char *p_s_new_alpha_filename, int n_new_texture_type, int n_new_width, int n_new_height, int n_new_depth)
	{
		if(p_s_filename)
			delete[] p_s_filename;
		if(p_s_alpha_filename)
			delete[] p_s_alpha_filename;
		p_s_filename = 0;
		p_s_alpha_filename = 0;
		//
		if(p_s_filename && !(p_s_filename = new char[strlen(p_s_new_filename) + 1]))
			return false;
		if(p_s_new_alpha_filename &&
		   !(p_s_alpha_filename = new char[strlen(p_s_new_alpha_filename) + 1]))
			return false;
		if(p_s_filename)
			strcpy(p_s_filename, p_s_new_filename);
		if(p_s_new_alpha_filename)
			strcpy(p_s_alpha_filename, p_s_new_alpha_filename);
		//
		n_references = 1;
		n_texture_type = n_new_texture_type;
		//
		n_size[0] = n_new_width;
		n_size[1] = n_new_height;
		n_size[2] = n_new_depth;

		return true;
	}
};

struct TBmp;

class /*__declspec(dllexport)*/ CTextureServer {
	std::vector<TTextureInfo> m_texture_list;
	int m_b_sorted;

public:
	CTextureServer();
	~CTextureServer();

	unsigned int Load_Texture_1D(CGLState *p_state, char *p_s_filename, GLenum n_format, int b_create_mipmaps, char *p_s_alpha_filename = 0);
	unsigned int Load_Texture_2D(CGLState *p_state, char *p_s_filename, GLenum n_format, int b_create_mipmaps, char *p_s_alpha_filename = 0);
	unsigned int Load_Texture_3D(CGLState *p_state, char *p_s_filename, int n_depth, GLenum n_format, int b_create_mipmaps, char *p_s_alpha_filename = 0);
	unsigned int Load_Texture_CubeMap(CGLState *p_state, char *p_s_filename, GLenum n_format, int b_create_mipmaps, char *p_s_alpha_filename = 0);

	unsigned int Create_Texture_1D(CGLState *p_state, int n_width, GLenum n_format, int b_create_mipmaps,
		void *p_data = 0, GLenum n_data_format = GL_RGBA, GLenum n_data_type = GL_UNSIGNED_BYTE);
	unsigned int Create_Texture_2D(CGLState *p_state, int n_width, int n_height, GLenum n_format, int b_create_mipmaps,
		void *p_data = 0, GLenum n_data_format = GL_RGBA, GLenum n_data_type = GL_UNSIGNED_BYTE);
	unsigned int Create_Texture_3D(CGLState *p_state, int n_width, int n_height, int n_depth, GLenum n_format, int b_create_mipmaps,
		void *p_data = 0, GLenum n_data_format = GL_RGBA, GLenum n_data_type = GL_UNSIGNED_BYTE);
	unsigned int Create_Texture_CubeMap(CGLState *p_state, int n_width, GLenum n_format, int b_create_mipmaps);

	void SortTextures();

	int n_Texture_Size(unsigned int n_ogl_id, int n_dimension) const;
	unsigned int Find_Texture(char *p_s_filename, unsigned int n_prev_texture = 0, char *p_s_alpha_filename = 0) const;

	int Use_Texture(unsigned int n_ogl_id); // increase reference counter
	int Delete_Texture(unsigned int n_ogl_id);

	void Delete_AllTextures();

protected:
	int Load_Bitmaps(char *p_s_filename, char *p_s_alpha_filename, TBmp &r_t_rgba_result);
	static int n_CompareTexInfo(const void *p_elem1, const void *p_elem2);
};

class /*__declspec(dllexport)*/ CTextureManager {
	static CTextureServer *m_p_texture_server;
	static int m_n_instances;
	//
	std::vector<unsigned int> m_local_texture_list;
	CGLState *m_p_state;

public:
	CTextureManager(CGLState *p_state);
	~CTextureManager();

	unsigned int Load_Texture_1D(char *p_s_filename, GLenum n_format, int b_create_mipmaps, char *p_s_alpha_filename = 0);
	unsigned int Load_Texture_2D(char *p_s_filename, GLenum n_format, int b_create_mipmaps, char *p_s_alpha_filename = 0);
	unsigned int Load_Texture_3D(char *p_s_filename, int n_depth, GLenum n_format, int b_create_mipmaps, char *p_s_alpha_filename = 0);
	unsigned int Load_Texture_CubeMap(char *p_s_filename, GLenum n_format, int b_create_mipmaps, char *p_s_alpha_filename = 0);

	unsigned int Create_Texture_1D(int n_width, GLenum n_format, int b_create_mipmaps,
		void *p_data = 0, GLenum n_data_format = GL_RGBA, GLenum n_data_type = GL_UNSIGNED_BYTE);
	unsigned int Create_Texture_2D(int n_width, int n_height, GLenum n_format, int b_create_mipmaps,
		void *p_data = 0, GLenum n_data_format = GL_RGBA, GLenum n_data_type = GL_UNSIGNED_BYTE);
	unsigned int Create_Texture_3D(int n_width, int n_height, int n_depth, GLenum n_format, int b_create_mipmaps,
		void *p_data = 0, GLenum n_data_format = GL_RGBA, GLenum n_data_type = GL_UNSIGNED_BYTE);
	unsigned int Create_Texture_CubeMap(int n_width, GLenum n_format, int b_create_mipmaps);

	unsigned int Find_Texture(char *p_s_filename, unsigned int n_prev_texture = 0, char *p_s_alpha_filename = 0);

	void SortTextures();

	int n_Texture_Size(unsigned int n_ogl_id, int n_dimension);

	int Use_Texture(unsigned int n_ogl_id);
	int Delete_Texture(unsigned int n_ogl_id);

	int Delete_LocalTextures();
	// delete texture / delete textures, loaded by this manager

	static const CTextureServer *p_Server();
};

#endif //__TEXTUREMAN_INCLUDED